
/*---------------------------------------------------------------------------
   
   File name    :       maskvoids.c
   Authors      :       J. Dietrich, T. Erben
   Created on   :       2006/2007
   Description  :       image masking by analysing object densities

 *--------------------------------------------------------------------------*/

/*
        $Id: maskvoids.c,v 1.7 2007-11-21 18:40:15 thomas Exp $
        $Author: thomas $
        $Date: 2007-11-21 18:40:15 $
        $Revision: 1.7 $
*/

/*---------------------------------------------------------------------------
                               History comments
 ---------------------------------------------------------------------------*/

/* 
   20.07.07:
   File included in masking codes

   12.09.2007:
   - The output FLAG image is now consistently treated as 8-bit BYTE
     image. As it contains only zeros one ones this accuracy is
     sufficient.
   - I speed up the calculation of the final FLAG image a bit.

   20.09.2007:
   I corrected a bug in the calculation of the size of the binned image.
   If objects lie at the 'upper right' corner of the original image
   the calculated size for binned images was too small by 1.

   21.09.2007:
   Another change concerning the bug from 20.09.2007   

   18.11.2007:
   - I included the possibility to write out several debug images
     (binned object distribution, estimated gradient field etc.).
     I included apropriate command line arguments.
   - I included program documentation

   19.11.2007:
   - I explicitely compile renamed versions of the GNU getopt function
     here. In this way we can use the 'getopt' function also on non-GNU
     systems.
*/

/*---------------------------------------------------------------------------
                                   Includes
 ---------------------------------------------------------------------------*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <fitsio.h>

#include "maskvoids.h"
#include "mygetopt.h"

/*---------------------------------------------------------------------------
                                    Main    
 ---------------------------------------------------------------------------*/

int main(int argc, char **argv) 
{
    int option;
    int status = 0;
    int hdutype;
    unsigned int m, n;  /* No. of pixels in x, y direction of ref img */
    unsigned int o, p;  /* No. of pixels in x, y of internal binned img */
    float *binimg, *gradimg;
    float thresh;
    long naxes[2];
    long nobj;          /* Number of objects */
    unsigned long i, i2;
    unsigned int j, k;
    float den, binfac;   /* Objects per binned pixel, binning factor */
    float tmpbin; 
    float *xg, *yg;
    char *incat = NULL;
    char *refimg = NULL;
    char *regfilename = NULL;
    char *xposkeyname = NULL;
    char *yposkeyname = NULL;
    char *cattablename = NULL;
    char *outimg;       /* The output mask image only contains zeros and one.
                           Hence, char is sufficient!
                        */
    fitsfile *cat, *img;

    FILE *regfile;

    /* names of optional output images */
    char *flagname = NULL;  /* name of the final output flag image */
    char *binname = NULL;  /* name of the binned object density image */
    char *limitedbinname = NULL; /* name of the binned object density image */
                                 /* with limited range */
    char *gradname = NULL;  /* name of the gradient image; gradients from */
                            /* object density */
    char *finalbinname = NULL;  /* 'final' image which is converted to output mask */

    /* default values */
    den = 4.0;
    thresh = 4.0;

    while ((option = getopt(argc, argv, "i:d:g:o:r:f:x:y:t:b:l:a:n:")) != -1 ) 
    {
        switch (option) 
        {
            case 'h': 
            {
                print_usage();
                exit(0);
                break;
            }
            case 'i': 
            {
                incat = strdup(optarg);
                break;
            }
            case 'd': 
            {
                den = atof(optarg);
                break;
            }
            case 'g': 
            {
                thresh = atof(optarg);
                break;
            }
            case 'r': 
            {
                refimg = strdup(optarg);
                break;
            }
            case 'f': 
            {
                regfilename = strdup(optarg);
                break;
            }
            case 'x': 
            {
                xposkeyname = strdup(optarg);
                break;
            }
            case 'y': 
            {
                yposkeyname = strdup(optarg);
                break;
            }
            case 't': 
            {
                cattablename = strdup(optarg);
                break;
            }
            case 'b': 
            {
                binname = strdup(optarg);
                break;
            }
            case 'l': 
            {
                limitedbinname = strdup(optarg);
                break;
            }
            case 'a': 
            {
                gradname = strdup(optarg);
                break;
            }
            case 'n': 
            {
                finalbinname = strdup(optarg);
                break;
            }
            case 'o': 
            {
                flagname = strdup(optarg);
                break;
            }
            default: 
            {
                fprintf(stderr, "Unknown option.\n");
                print_usage();
                exit(EXIT_FAILURE);
            }
        }
    }

    /* sanity checks */
    if (!incat || !refimg) 
    {
        fprintf(stderr, "Need input catalog and reference image!\n");
        print_usage();
        exit(EXIT_FAILURE);
    }
 
    if((regfilename == NULL) && (flagname == NULL))
    {
        fprintf(stderr, "Nothing to do!! Exiting\n");
        print_usage();
        exit(EXIT_FAILURE);
    }

    if(xposkeyname == NULL)
    {
        xposkeyname = strdup("Xpos");
    }

    if(yposkeyname == NULL)
    {
        yposkeyname = strdup("Ypos");
    }

    if(cattablename == NULL)
    {
        cattablename = strdup("OBJECTS");
    }

    /* program execution starts here */
    fits_open_file(&cat, incat, READONLY, &status);
    fits_movnam_hdu(cat, BINARY_TBL, cattablename, 0, &status);
    if (status == BAD_HDU_NUM) 
    {
        status = 0;
        fits_movnam_hdu(cat, ASCII_TBL, cattablename, 0, &status);
    }
    if (status) 
    {
        fits_report_error(stderr, status);
        exit(EXIT_FAILURE);
    }

    fits_get_num_rows(cat, &nobj, &status);
    if (status) 
    {
        fits_report_error(stderr, status);
        exit(EXIT_FAILURE);
    }
    
    xg = calloc(nobj, sizeof(float));
    yg = calloc(nobj, sizeof(float));
    if (!xg || !yg) 
    {
        fprintf(stderr, "Allocation of galaxy position vectors failed.\n");
        exit(EXIT_FAILURE);
    }
    readfitscol(cat, xposkeyname, nobj, xg);
    readfitscol(cat, yposkeyname, nobj, yg);
    fits_close_file(cat, &status);
    cat = NULL;
    printf("Read %ld object positions.\n", nobj);
    
    fits_open_file(&img, refimg, READONLY, &status);
    if (fits_movabs_hdu(img, 1, &hdutype, &status) || hdutype != IMAGE_HDU )
    {
        fprintf(stderr, "Reference image is not an image.\n");
        exit(EXIT_FAILURE);
    }
    fits_read_key(img, TINT, "NAXIS1", &m, NULL, &status);
    fits_read_key(img, TINT, "NAXIS2", &n, NULL, &status);
    if (status) 
    {
        fits_report_error(stderr, status);
        exit(EXIT_FAILURE);
    }
    printf("Reference image is %u x %u pix^2.\n", m, n);
    fits_close_file(img, &status);

    binfac = sqrt(nobj / (den * m * n));

    /* The additional '+ 1.0' is to account for fractional values
       of 'm(n) * binfac.'. We need to make sure to get to the
       next larger integer to avoid writing out of array bounds
       later.
    */
    o = (unsigned int)floor((m * binfac) + 0.5 + 1.0);
    p = (unsigned int)floor((n * binfac) + 0.5 + 1.0);

    printf("The binning factor is: %f\n", binfac);
    printf("Internal image is %u x %u pix^2.\n", o, p);

    binimg = (float *)calloc(o * p, sizeof(float));
    if (!binimg) 
    {
        fprintf(stderr, "Allocation of binned image failed.\n");
        exit(EXIT_FAILURE);
    }
    naxes[0] = o;
    naxes[1] = p;
    
    make_binimg(binimg, nobj, binfac, xg, yg, naxes);

    free(xg);
    free(yg);

    if(binname != NULL)
    {
      /* the following remove is in case the named image exists
         already. In this case the CFITIS library would not overwrite
         the old image but abort the program 
      */
      remove(binname);
      save_float_image(binname, binimg, o, p);
    }

    smooth(binimg, naxes);

    lim_range(binimg, naxes, den);

    if(limitedbinname != NULL)
    {
      remove(limitedbinname);
      save_float_image(limitedbinname, binimg, o, p);
    }

    gradimg = calloc(o * p, sizeof(float));
    if (!gradimg) 
    {
        fprintf(stderr, "Allocation of grad image failed.\n");
        exit(EXIT_FAILURE);
    }
    memcpy(gradimg, binimg, o * p * sizeof(float));

    grad(gradimg, naxes);
    
    if(gradname != NULL)
    {
      remove(gradname);
      save_float_image(gradname, gradimg, o, p);
    }

    for (j = 0; j < o; j++) 
    {
        for (k = 0; k < p; k++) 
        {
            i = j + k*o;
            if (binimg[i] == 0.0)
            {
                binimg[i] = 1.0;
            }
            else
            {
                binimg[i] = 0.0;
            }
            gradimg[i] = gradimg[i] > thresh ? 1.0 : 0.0;
            binimg[i] = (int)gradimg[i] | (int)binimg[i];
    
        }
    }
    free(gradimg);

    smooth(binimg, naxes);

    for (j=0; j<o; j++) 
    {
        for (k=0; k<p; k++) 
        {
            i = j + k * o;
            if (binimg[i] > 0.0)
            {
                binimg[i] = 1.0;
            }
        }
    }

    if(finalbinname != NULL)
    {
      remove(finalbinname);
      save_float_image(finalbinname, binimg, o, p);
    }

    /* write region file */
    if(regfilename != NULL)
    {
        if((regfile=fopen(regfilename, "w"))==NULL)
        {
            printf("Could not open file %s!!\n", regfilename);
            exit(1);
        }

        printf("writing regions to %s\n", regfilename);
        /* the file must start with three arbitrary comment
           lines to be a valid saoimage/ds9 region file 
        */
        fprintf(regfile, "#\n#\n#\n");

        i = 0;
        for (k = 0; k < p; k++) 
        {
            for (j = 0; j < o; j++) 
            {
                if (fabs(binimg[i++] - 1.0) < 1.0e-06)
                {
                    /* around each marked point in the binned 
                       image we create a square in the unbinned 
                       one. The square side length is just the 
                       binsize. 
                       Also give, on each corner, two pixels
                       margin
                    */ 
                    fprintf(regfile, " POLYGON(%d,%d,%d,%d,%d,%d,%d,%d,%d,%d)\n",
                       (int)floor(j/binfac-1/(2.*binfac))-2, 
                       (int)floor(k/binfac-1/(2.*binfac))-2, 
                       (int)floor(j/binfac+1/(2.*binfac))+2, 
                       (int)floor(k/binfac-1/(2.*binfac))-2, 
                       (int)floor(j/binfac+1/(2.*binfac))+2, 
                       (int)floor(k/binfac+1/(2.*binfac))+2, 
                       (int)floor(j/binfac-1/(2.*binfac))-2, 
                       (int)floor(k/binfac+1/(2.*binfac))+2, 
                       (int)floor(j/binfac-1/(2.*binfac))-2, 
                       (int)floor(k/binfac-1/(2.*binfac))-2);
                }
            }
        }
        fclose(regfile);
        printf("region file %s written\n", regfilename);
    }
    
    /* save flag image */
    if (flagname != NULL)
    {
        printf("writing flag image to %s\n", flagname);

        outimg = calloc(m * n, sizeof(char));

        if (!outimg) 
        {       
            fprintf(stderr, "Allocation of output image failed.\n");
            exit(EXIT_FAILURE);
        }
        
        /* Recompute to fix round-off errors */
        binfac = ((float)p-1)/((float)n-1);
        
        i = 0;
        for (k = 0; k < n; k++) 
        {
            tmpbin = floor(k * binfac) * o;
            for (j = 0; j < m; j++) 
            {
                i2 = (unsigned long)(floor(j * binfac) + tmpbin);
                if (fabs(binimg[i2] - 1.0) < 1.0e-06)
                {
                    outimg[i] = 1;
                }
                i++;
            }
        }

        remove(flagname);
        naxes[0] = m;
        naxes[1] = n;

        fits_create_file(&img, flagname, &status);
        fits_create_img(img, BYTE_IMG, 2, naxes, &status);

        if (status) 
        {
            fits_report_error(stderr, status);
            exit(EXIT_FAILURE);
        }
        
        fits_write_img(img, TBYTE, 1, m*n, outimg, &status);
        if (status) 
        {
            fits_report_error(stderr, status);
            exit(EXIT_FAILURE);
        }

        fits_close_file(img, &status);
        if (status) 
        {
            fits_report_error(stderr, status);
            exit(EXIT_FAILURE);
        }
        
        free(outimg);
        printf("flag image %s written\n", flagname);
    }

    /* free allocated memory */
    free(binimg);
    free(incat);
    free(refimg);
    free(xposkeyname);
    free(yposkeyname);
    free(cattablename);

    if(flagname != NULL)       { free(flagname); }
    if(regfilename != NULL)    { free(regfilename); }
    if(binname != NULL)        { free(binname); }
    if(gradname != NULL)       { free(gradname); }
    if(limitedbinname != NULL) { free(finalbinname); }

    printf("All Done!\n");

    return EXIT_SUCCESS;
}


void lim_range(float *img, long *naxes, float den) 
{
    unsigned long i;
    long j, k;

    for (j = 0; j < naxes[0]; j++) 
    {
        for (k = 0; k < naxes[1]; k++) 
        {
            i = j + k * naxes[0];
            if (img[i] > 0.75 * den) 
            {
                img[i] = den;
            }
            else 
            {
                if (img[i] < den / 2.)
                {
                    img[i] = 0;
                }
            }
        }
    }
}



void grad(float *img, long *naxes) 
{
    /* Sobel masks for gradient computation in x and y direction */
    float Gx[3][3] = {{-1, 0, 1},
                      {-2, 0, 2},
                      {-1, 0, 1}};
    float Gy[3][3] = {{1, 2, 1},
                      {0, 0, 0},
                      {-1, -2, -1}};
    float *gximg, *gyimg;
    float x;
    unsigned long i, i2;
    int j, k, l, m;

    gximg = calloc(naxes[0]*naxes[1], sizeof(float));
    gyimg = calloc(naxes[0]*naxes[1], sizeof(float));
    if (!gximg || !gyimg ) 
    {
        fprintf(stderr, "Allocation of gximg/gyimg failed.\n");
        exit(EXIT_FAILURE);
    }

    /* Compute the gradient components by convolution with the Sobel masks */
    for (j=0; j<naxes[0]; j++) 
    {
        for (k=0; k<naxes[1]; k++) 
        {
            for (l=-1; l<2; l++) 
            {
                for (m=-1; m<2; m++) 
                {
                    if (((j+l)>=0) && ((j+l)<naxes[0]) && 
                        ((k+m)>=0) && ((k+m)<naxes[1])) 
                    {
                        i = j+k*naxes[0];
                        i2 = (j+l)+(k+m)*naxes[0];
                        gximg[i] += img[i2]*Gx[l+1][m+1]; 
                        gyimg[i] += img[i2]*Gy[l+1][m+1]; 
                    }
                }
            }
        }
    }
    /* Compute the absolute value of the gradient */
    for (j=0; j<naxes[0]; j++) 
    {
        for (k=0; k<naxes[1]; k++) 
        {
            i = j+k*naxes[0];
            x = sqrt(SQR(gximg[i]) + SQR(gyimg[i]));
            img[i] = x;
        }
    }
    free(gximg);
    free(gyimg);
}


void make_binimg(float *img, unsigned long nobj, float binfac, 
                         float *x, float *y, long *naxes) 
{
    unsigned long i;
    int j, k;

    for (i = 0; i < nobj; i++) 
    {
        j = (unsigned int)floor((x[i] * binfac) + 0.5);
        k = (unsigned int)floor((y[i] * binfac) + 0.5);
        img[j + k * naxes[0]] += 1;
    }
}


void smooth(float *img, long *naxes) 
{
    float *tmpimg;
    int j, k, l, m;
    /* An unnormalized Gaussian with FWHM 1.4 
     * Divide by 159 for normalization         */
    float gauss[5][5] = {{2, 4, 5, 4, 2},
                         {4, 9, 12, 9, 4},
                         {5, 12, 15, 12, 5},
                         {4, 9, 12, 9, 4},
                         {2, 4, 5, 4, 2}};

    tmpimg = calloc(naxes[0]*naxes[1], sizeof(float));
    if (!tmpimg) 
    {
        fprintf(stderr, "Allocation of tmp image failed.\n");
        exit(EXIT_FAILURE);
    }
    memcpy(tmpimg, img, naxes[0]*naxes[1]*sizeof(float));

    /* The convolution with the Gauss matrix */
    for (j=0; j<naxes[0]; j++) 
    {
        for (k=0; k<naxes[1]; k++) 
        {
            for (l=-2; l<3; l++) 
            {
                for (m=-2; m<3; m++) 
                {
                    if (((j+l)>=0) && ((j+l)<naxes[0]) && 
                        ((k+m)>=0) && ((k+m)<naxes[1]))
                    { 
                        img[j+k*naxes[0]] += 
                        tmpimg[(j+l)+(k+m)*naxes[0]]*gauss[l+2][m+2]/159.; 
                    }
                }
            }
        }
    }
    free(tmpimg);
}


void readfitscol(fitsfile *fptr, char *colnam, long nrows, float *arr) 
{
    int colnum;
    int status = 0;
    
    fits_get_colnum(fptr, CASESEN, colnam, &colnum, &status);
    if (status) 
    {
        fits_report_error(stderr, status);
        exit(EXIT_FAILURE);
    }
    fits_read_col(fptr, TFLOAT, colnum, 1, 1, nrows, NULL, arr, 0, &status);
}

/* save a float array to a FITS image */
void save_float_image(char *aname, float *adata, int am, int an)
{
    int status = 0;
    long naxes[2];
    fitsfile *img;

    naxes[0] = am;
    naxes[1] = an;

    fits_create_file(&img, aname, &status);
    fits_create_img(img, FLOAT_IMG, 2, naxes, &status);

    if (status) 
    {
        fits_report_error(stderr, status);
        exit(EXIT_FAILURE);
    }
    
    fits_write_img(img, TFLOAT, 1, am * an, adata, &status);
    if (status) 
    {
        fits_report_error(stderr, status);
        exit(EXIT_FAILURE);
    }

    fits_close_file(img, &status);
    if (status) 
    {
        fits_report_error(stderr, status);
        exit(EXIT_FAILURE);
    }
    
}

/*
 * This function only gives the usage message for the program
 */

void print_usage(void)
{
    fprintf(stdout, "PROGRAMNAME\n");
    fprintf(stdout, "        maskvoids - mask images by analysing object number densities\n");
    fprintf(stdout, "\n");
    fprintf(stdout, "SYNOPSIS\n");
    fprintf(stdout, "        maskvoids -i -r [-f, -o] [-d, -g, -x, -y, -t, -b, -l, -a, -n]\n");
    fprintf(stdout, "                -i      # input catalogue\n");
    fprintf(stdout, "                -r      # reference image\n");
    fprintf(stdout, "              [ -o      # name of output flag map                       ]\n");
    fprintf(stdout, "              [ -d      # mean number of objects per box (default: 4.0) ]\n");
    fprintf(stdout, "              [ -g      # gradient threshhold            (default: 4.0) ]\n");
    fprintf(stdout, "              [ -f      # name of output region file                    ]\n");
    fprintf(stdout, "              [ -t      # object table name (default: OBJECTS)          ]\n");
    fprintf(stdout, "              [ -x      # name of x coordinate (default: Xpos)          ]\n");
    fprintf(stdout, "              [ -y      # name of y coordinate (default: Ypos)          ]\n");
    fprintf(stdout, "              [ -b      # name of object density binned image           ]\n");
    fprintf(stdout, "              [         # (default: not created)                        ]\n");
    fprintf(stdout, "              [ -l      # name of object density binned image           ]\n");
    fprintf(stdout, "              [         # with limited value range                      ]\n");
    fprintf(stdout, "              [         # (default: not created)                        ]\n");
    fprintf(stdout, "              [ -a      # name of gradient density image of object      ]\n");
    fprintf(stdout, "              [         # distribution                                  ]\n");
    fprintf(stdout, "              [         # (default: not created)                        ]\n");
    fprintf(stdout, "              [ -n      # name of final binned image which is           ] \n");
    fprintf(stdout, "              [         # translated to output flag                     ]\n");
    fprintf(stdout, "              [         # (default: not created)                        ]\n");
    fprintf(stdout, "\n");
    fprintf(stdout, "        either '-f' or '-o' (or both) must be given!\n");
    fprintf(stdout, "\n");
    fprintf(stdout, "DESCRIPTION\n");
    fprintf(stdout, "   The program creates object masks (either FLAG images for the use with\n");
    fprintf(stdout, "   SExtractor or saoimage/ds9 region files or both) by considering the \n");
    fprintf(stdout, "   two-dimenasional object distribution in the input LDAC catalogue\n");
    fprintf(stdout, "   (option '-i'). The determination of regions that should be masked\n");
    fprintf(stdout, "   is done in the following steps:\n");
    fprintf(stdout, "\n");
    fprintf(stdout, "   1. All objects in the input cataloue are put into cells of a grid\n");
    fprintf(stdout, "      whose size is chosen such that the average number of objects per\n");
    fprintf(stdout, "      cell is the number 'd' given in the '-d' (density) argument. \n");
    fprintf(stdout, "      The area available to objects is obtained by the size (NAXIS1,\n");
    fprintf(stdout, "      NAXIS2 header keywords of the reference image (option '-r'))\n");
    fprintf(stdout, "   2. The grid is smoothed with a Gaussian kernel with a FWHM of 1.4\n");
    fprintf(stdout, "      pixels. We call the output of this step 'S'.\n");
    fprintf(stdout, "   3. The dynamic range os 'S' is limited. All pixels in 'S' with a\n");
    fprintf(stdout, "      value of < 0.5*'d' are set to zero; all pixels with a value >\n");
    fprintf(stdout, "      0.75 * 'd' are set to 'd'. The zeros will later enter the mask    \n");
    fprintf(stdout, "      in any case. The resulting array is called 'D'.\n");
    fprintf(stdout, "   4. An array containing the absolute values of the gradients in 'D'\n");
    fprintf(stdout, "      is calculated and labeled 'G'.  All pixels exceeding a\n");
    fprintf(stdout, "      threshhold in this gradient (option '-g') enter the final mask\n");
    fprintf(stdout, "      as well. This step catches regions with sudden, rapid changes in\n");
    fprintf(stdout, "      object density.  Note that the maximum possible value for the\n");
    fprintf(stdout, "      resulting derivatives is sqrt(2)*4*'d'!\n");
    fprintf(stdout, "   5. A first flag array is created. It contains ones for all pixels\n");
    fprintf(stdout, "      that need to be masked up to now and one everywhere else.\n");
    fprintf(stdout, "   6. The resulting array is smoothed again with the Gaussian of 1.4\n");
    fprintf(stdout, "      pixels FWHM and the resulting array is called 'F'. \n");
    fprintf(stdout, "      This new 'blurring' is done to account for the fact\n");
    fprintf(stdout, "      that the initial smmothing shrinked the area not ceovered by\n");
    fprintf(stdout, "      objects.\n");
    fprintf(stdout, "   7. All non-zero pixels after the last step enter the final FLAG\n");
    fprintf(stdout, "      image and/or region file.\n");
    fprintf(stdout, "\n");
    fprintf(stdout, "   Check images (FITS format) of the arrays 'S', 'D', 'G' and 'F' can\n");
    fprintf(stdout, "   be created with command line options '-b', '-l', '-a' and '-n'\n");
    fprintf(stdout, "   respectively.\n");
    fprintf(stdout, "   \n");
    fprintf(stdout, "   The program efficiently picks large-scale features modifying\n");
    fprintf(stdout, "   significantly the object distribution. To work efficiently, a very\n");
    fprintf(stdout, "   dense input catalogue (low detection threshold and low minimum area\n");
    fprintf(stdout, "   as SExtractor arguments) should be given.\n");
    fprintf(stdout, "\n");
    fprintf(stdout, "AUTHOR\n");
    fprintf(stdout, "        Joerg Dietrich (jdietric@eso.org)\n");
    fprintf(stdout, "        (original program version with name 'make_mask')\n");
    fprintf(stdout, "\n");
    fprintf(stdout, "        extended by\n");
    fprintf(stdout, "        Thomas Erben: terben@astro.uni-bonn.de\n");
}
